//	Flower.cpp - The basic flower class.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "drawing/Drawer.h"
#include "GamePlayState.h"
#include "ColourHelpers.h"
#include "Random.h"
#include "Flower.h"

//------------------------------------------------------------------------------
Flower::Flower(const TwoFloats& pos, bool brandNew):
position(pos),
ringCount(0.0f),
seedCount(-1000.0f)
{
	if(brandNew)
	{
		life = 0.0f;
		synthIndex = 0;
	}
	else
	{
		life = Random::getGlobal().getRandomFloat();
		synthIndex = Random::getGlobal().getRandomInt(8);
	}

	colour.x = Random::getGlobal().getRandomFloat();
	colour.y = 0.5f;
	colour.z = 0.85f;
	colour = ColourHelpers::hslToRgb(colour);
}

//------------------------------------------------------------------------------
Flower::~Flower()
{

}

//------------------------------------------------------------------------------
bool Flower::update(float delta)
{
	bool retval = false;

	if(ringCount > 0.0f)
		ringCount -= delta;

	life += 0.001f * delta;
	if(life >= 1.0f)
	{
		++synthIndex;
		if(synthIndex > 7)
			retval = true;

		life = 0.0f;
	}

	if(seedCount > 0.0f)
		seedCount -= 0.005f * delta;

	return retval;
}

//------------------------------------------------------------------------------
void Flower::draw(shared_ptr<Drawer>& d, const TwoFloats& topLeft)
{
	const float size = 0.2f + (((float)synthIndex/8.0f) * 0.8f);

	if(position.isWithin(topLeft-64.0f, d->getDrawingAreaSize()+128.0f))
	{
		d->drawImage(L"Flower",
					 position-topLeft-(64.0f*size),
					 size,
					 colour);
	}
}

//------------------------------------------------------------------------------
bool Flower::collide(GamePlayState& state)
{
	bool retval = false;

	if(ringCount <= 0.0f)
	{
		retval = true;
		ringCount = 90.0f;
	}

	if((seedCount > -1000.0f) && (seedCount <= 0.0f))
	{
		state.addSeed(position);
		seedCount = -1000.0f;
	}
	else if(seedCount <= -1000.0f)
	{
		if((synthIndex > 2) && (synthIndex < 6)) //Only emit seeds from adult (but not too adult) flowers.
			seedCount = 4.0f + (Random::getGlobal().getRandomFloat() * 8.0f);
	}

	return retval;
}

//------------------------------------------------------------------------------
float Flower::getRadius() const
{
	const float size = 0.2f + (((float)synthIndex/8.0f) * 0.8f);

	return 64.0f * size;
}
