//	GamePlayState.cpp - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "PastarenaSounder.h"
#include "drawing/Drawer.h"
#include "GamePlayState.h"
#include "ColourHelpers.h"
#include "PlayerBullet.h"
#include "FlowerRing.h"
#include "MiscHelper.h"
#include "Flower.h"
#include "Random.h"
#include "Seed.h"
#include "Bee.h"

//------------------------------------------------------------------------------
GameStateCreator<GamePlayState> GamePlayState::creator;

//------------------------------------------------------------------------------
GamePlayState::GamePlayState(Application& app, SounderPtr& sound):
GameState(app, sound),
sounder(dynamic_pointer_cast<PastarenaSounder>(sound))
{
	int i, j;
	ThreeFloats tempHSL;
	std::vector<Biome> availableBiomes;
	
	tempHSL.x = Random::getGlobal().getRandomFloat();
	tempHSL.y = 0.25f;
	tempHSL.z = 0.66f;
	for(i=0;i<sounder->getNumPitchSets();++i)
	{
		Biome b;

		b.colour = ColourHelpers::hslToRgb(tempHSL);
		b.pitchSet = i;

		availableBiomes.push_back(b);

		tempHSL.x += 1.0f/(float)sounder->getNumPitchSets();
		if(tempHSL.x > 1.0f)
			tempHSL.x -= 1.0f;
	}

	for(i=0;i<NumBiomes;++i)
	{
		for(j=0;j<NumBiomes;++j)
			biomes[i][j] = availableBiomes[Random::getGlobal().getRandomInt(availableBiomes.size())];
	}

	generateFlowers();

	/*for(i=0;i<NumBees;++i)
	{
		bees.push_back(shared_ptr<Bee>(new Bee(TwoFloats(Random::getGlobal().getRandomFloat() * (1280.0f*20.0f),
											   Random::getGlobal().getRandomFloat() * (1280.0f*20.0f)))));
	}*/

	setMousePos(getDrawingAreaSize() * 0.5f);
}

//------------------------------------------------------------------------------
GamePlayState::~GamePlayState()
{

}

//------------------------------------------------------------------------------
void GamePlayState::update()
{
	int i;
	const float delta = 1.0f;

	player.update(delta, *this);

	for(i=(playerBullets.size()-1);i>=0;--i)
	{
		if(playerBullets[i]->update(delta))
			playerBullets.erase(playerBullets.begin()+i);
	}

	/*for(i=0;i<bees.size();++i)
		bees[i]->update(delta, *this);*/

	for(i=(flowers.size()-1);i>=0;--i)
	{
		if(flowers[i]->update(delta))
		{
			particles.trigger(flowers[i]->getPosition());
			flowers.erase(flowers.begin()+i);
		}
	}

	for(i=(flowerRings.size()-1);i>=0;--i)
	{
		if(flowerRings[i]->update(delta))
			flowerRings.erase(flowerRings.begin()+i);
	}

	for(i=(seeds.size()-1);i>=0;--i)
	{
		if(seeds[i]->update(delta))
		{
			flowers.push_back(shared_ptr<Flower>(new Flower(seeds[i]->getPosition(), true)));

			particles.trigger(seeds[i]->getPosition());
			seeds.erase(seeds.begin()+i);
		}
	}

	particles.update(delta);

	collideBulletsFlowers();
	collideRingsFlowers();
	//collideBeesFlowers();
}

//------------------------------------------------------------------------------
void GamePlayState::draw(DrawerPtr& d, float timestep)
{
	unsigned int i;
	ThreeFloats col;
	const TwoFloats topLeft(player.getPosition() - (d->getDrawingAreaSize()*0.5f));

	col = biomes[getBiomeX(player.getPosition())][getBiomeY(player.getPosition())].colour;
	d->fillRect(0.0f, d->getDrawingAreaSize(), col);

	for(i=0;i<playerBullets.size();++i)
		playerBullets[i]->draw(d, topLeft);

	particles.draw(d, topLeft);

	for(i=0;i<flowers.size();++i)
		flowers[i]->draw(d, topLeft);

	for(i=0;i<flowerRings.size();++i)
		flowerRings[i]->draw(d, topLeft);

	for(i=0;i<seeds.size();++i)
		seeds[i]->draw(d, topLeft);

	/*for(i=0;i<bees.size();++i)
		bees[i]->draw(d, topLeft);*/

	player.draw(d);
}

//------------------------------------------------------------------------------
void GamePlayState::keyDown(const std::wstring& action,
							const std::wstring& keyPressed,
							int keyModifier)
{
	if(action == L"Escape")
		quit();
}

//------------------------------------------------------------------------------
void GamePlayState::keyUp(const std::wstring& action,
						  const std::wstring& keyPressed,
						  int keyModifier)
{
	
}

//------------------------------------------------------------------------------
void GamePlayState::mouseMove(const TwoFloats& pos)
{
	const TwoFloats centre(getDrawingAreaSize() * 0.5f);

	player.move(pos-centre);
	setMousePos(centre);
}

//------------------------------------------------------------------------------
void GamePlayState::addPlayerBullet(const TwoFloats& pos, const TwoFloats& dir)
{
	float tempf;
	const TwoFloats origin;

	playerBullets.push_back(shared_ptr<PlayerBullet>(new PlayerBullet(pos, dir)));

	tempf = origin.getAngle(dir);
	tempf += NIALL_PI;
	if(tempf > (2.0f*NIALL_PI))
		tempf -= (2.0f*NIALL_PI);
	else if(tempf < 0.0f)
		tempf += NIALL_PI;
	tempf /= (2.0f*NIALL_PI);
	sounder->triggerHi(tempf,
					   biomes[getBiomeX(player.getPosition())][getBiomeY(player.getPosition())].pitchSet);
}

//------------------------------------------------------------------------------
void GamePlayState::addSeed(const TwoFloats& pos)
{
	float tempf;
	const TwoFloats origin;
	const TwoFloats topLeft(player.getPosition() - (getDrawingAreaSize()*0.5f));

	seeds.push_back(shared_ptr<Seed>(new Seed(pos)));

	if(pos.isWithin(topLeft, getDrawingAreaSize()))
	{
		tempf = origin.getAngle(seeds.back()->getDirection());
		tempf += NIALL_PI;
		if(tempf > (2.0f*NIALL_PI))
			tempf -= (2.0f*NIALL_PI);
		else if(tempf < 0.0f)
			tempf += NIALL_PI;
		tempf /= (2.0f*NIALL_PI);
		sounder->triggerHi(tempf,
						   biomes[getBiomeX(pos)][getBiomeY(pos)].pitchSet);
	}
}

//------------------------------------------------------------------------------
void GamePlayState::generateFlowers()
{
	int i, j;
	int numFlowers;
	TwoFloats position;
	TwoFloats direction;
	const int NumBunches = 1024;

	for(i=0;i<NumBunches;++i)
	{
		numFlowers = 1 + Random::getGlobal().getRandomInt(8);
		position.x = Random::getGlobal().getRandomFloat() * 1280.0f * 20.0f;
		position.y = Random::getGlobal().getRandomFloat() * 1280.0f * 20.0f;
		for(j=0;j<numFlowers;++j)
		{
			flowers.push_back(shared_ptr<Flower>(new Flower(position)));

			direction.x = (Random::getGlobal().getRandomFloat() * 2.0f) - 1.0f;
			direction.y = (Random::getGlobal().getRandomFloat() * 2.0f) - 1.0f;
			direction = direction.normal();

			position += direction * 128.0f;
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::collideBulletsFlowers()
{
	int i, j;
	float pan;
	float bulletRad;
	TwoFloats bulletPos;

	for(i=(playerBullets.size()-1);i>=0;--i)
	{
		bulletPos = playerBullets[i]->getPosition();
		bulletRad = playerBullets[i]->getRadius();

		for(j=0;j<(int)flowers.size();++j)
		{
			if(bulletPos.getDistance(flowers[j]->getPosition()) < (bulletRad+flowers[j]->getRadius()))
			{
				if(flowers[j]->collide(*this))
				{
					flowerRings.push_back(shared_ptr<FlowerRing>(new FlowerRing(flowers[j]->getPosition(),
																				flowers[j]->getColour())));

					pan = (flowers[j]->getPosition().x - player.getPosition().x)/380.0f;
					if(pan < -1.0f)
						pan = -1.0f;
					else if(pan > 1.0f)
						pan = 1.0f;
					pan = (pan+1.0f)*0.5f;
					sounder->triggerLo(flowers[j]->getSynthIndex(),
									   biomes[getBiomeX(flowers[j]->getPosition())][getBiomeY(flowers[j]->getPosition())].pitchSet,
									   player.getPosition().getDistance(flowers[j]->getPosition()),
									   pan);
				}

				particles.trigger(playerBullets[i]->getPosition());

				playerBullets.erase(playerBullets.begin()+i);
				break;
			}
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::collideRingsFlowers()
{
	float pan;
	float dist;
	float ringRad;
	unsigned int i, j;
	TwoFloats ringPos;
	float lower, upper;

	for(i=0;i<flowerRings.size();++i)
	{
		ringPos = flowerRings[i]->getPosition();
		ringRad = flowerRings[i]->getRadius();
		for(j=0;j<flowers.size();++j)
		{
			lower = ringRad + flowers[j]->getRadius() - 2.5f;
			upper = lower + 5.0f;
			dist = ringPos.getDistance(flowers[j]->getPosition());

			if((dist > lower) && (dist < upper))
			{
				if(flowers[j]->collide(*this))
				{
					flowerRings.push_back(shared_ptr<FlowerRing>(new FlowerRing(flowers[j]->getPosition(),
																				flowers[j]->getColour())));
					
					pan = (flowers[j]->getPosition().x - player.getPosition().x)/380.0f;
					if(pan < -1.0f)
						pan = -1.0f;
					else if(pan > 1.0f)
						pan = 1.0f;
					pan = (pan+1.0f)*0.5f;
					sounder->triggerLo(flowers[j]->getSynthIndex(),
									   biomes[getBiomeX(flowers[j]->getPosition())][getBiomeY(flowers[j]->getPosition())].pitchSet,
									   player.getPosition().getDistance(flowers[j]->getPosition()),
									   pan);
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::collideBeesFlowers()
{
	int i, j;
	float pan;
	float beeRad;
	TwoFloats beePos;

	for(i=(bees.size()-1);i>=0;--i)
	{
		beePos = bees[i]->getPosition();
		beeRad = bees[i]->getRadius();

		for(j=0;j<(int)flowers.size();++j)
		{
			if(beePos.getDistance(flowers[j]->getPosition()) < (beeRad+flowers[j]->getRadius()))
			{
				if(flowers[j]->collide(*this))
				{
					flowerRings.push_back(shared_ptr<FlowerRing>(new FlowerRing(flowers[j]->getPosition(),
																				flowers[j]->getColour())));

					pan = (flowers[j]->getPosition().x - player.getPosition().x)/380.0f;
					if(pan < -1.0f)
						pan = -1.0f;
					else if(pan > 1.0f)
						pan = 1.0f;
					pan = (pan+1.0f)*0.5f;
					sounder->triggerLo(flowers[j]->getSynthIndex(),
									   biomes[getBiomeX(flowers[j]->getPosition())][getBiomeY(flowers[j]->getPosition())].pitchSet,
									   player.getPosition().getDistance(flowers[j]->getPosition()),
									   pan);
				}

				break;
			}
		}
	}
}

//------------------------------------------------------------------------------
int GamePlayState::getBiomeX(const TwoFloats& val) const
{
	return (int)((val.x/(1280.0f*20.0f))*(float)NumBiomes);
}

//------------------------------------------------------------------------------
int GamePlayState::getBiomeY(const TwoFloats& val) const
{
	return (int)((val.y/(1280.0f*20.0f))*(float)NumBiomes);
}
