//	GamePlayState.h - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GAMEPLAYSTATE_H_
#define GAMEPLAYSTATE_H_

#include "ParticleGenerator.h"
#include "ThreeFloats.h"
#include "GameState.h"

#include "Player.h"

#include <vector>

class PastarenaSounder;
class PlayerBullet;
class FlowerRing;
class Flower;
class Seed;
class Bee;

///	The main gameplay state.
class GamePlayState : public GameState
{
  public:
	///	Constructor.
	GamePlayState(Application& app, SounderPtr& sound);
	///	Destructor.
	~GamePlayState();

	///	Called frequently to update the state.
	void update();
	///	Called to draw the state.
	void draw(DrawerPtr& d, float timestep);

	///	Called when a key down event is received.
	void keyDown(const std::wstring& action,
				 const std::wstring& keyPressed,
				 int keyModifier);
	///	Called when a key up event is received.
	void keyUp(const std::wstring& action,
			   const std::wstring& keyPressed,
			   int keyModifier);

	///	Used to test connected line drawing.
	void mouseMove(const TwoFloats& pos);

	///	Adds a player bullet.
	void addPlayerBullet(const TwoFloats& pos, const TwoFloats& dir);
	///	Adds a seed from a flower.
	void addSeed(const TwoFloats& pos);

	///	Returns the flowers in the game.
	const std::vector<shared_ptr<Flower> >& getFlowers() const {return flowers;};

	///	Returns the name of this state.
	const std::wstring getName() const {return L"GamePlayState";};
	///	Used to construct instances of this state.
	static GameStateCreator<GamePlayState> creator;
  private:
	///	Helper method. Populates the meadow with flowers.
	void generateFlowers();
	///	Helper method. Handles any PlayerBullet->Flower collisions.
	void collideBulletsFlowers();
	///	Helper method. Handles any FlowerRing->Flower collisions.
	void collideRingsFlowers();
	///	Helper method. Handles any Bee->Flower collisions.
	void collideBeesFlowers();
	///	Helper method. Returns the biome x-coord for the passed-in position.
	int getBiomeX(const TwoFloats& val) const;
	///	Helper method. Returns the biome y-coord for the passed-in position.
	int getBiomeY(const TwoFloats& val) const;

	///	Struct representing a biome.
	struct Biome
	{
		///	The biome's colour.
		ThreeFloats colour;
		///	The biome's pitch set.
		int pitchSet;
	};

	///	The number of biomes in the world (just go with a grid structure for now).
	enum
	{
		NumBiomes = 5,
		NumBees = 128
	};

	///	The current background colour.
	ThreeFloats backgroundColour;

	///	The player.
	Player player;
	///	Any active PlayerBullets.
	std::vector<shared_ptr<PlayerBullet> > playerBullets;

	///	All the flowers in the game.
	std::vector<shared_ptr<Flower> > flowers;
	///	Any active FlowerRings.
	std::vector<shared_ptr<FlowerRing> > flowerRings;
	///	Any active Seeds.
	std::vector<shared_ptr<Seed> > seeds;
	///	All the bees in the game.
	std::vector<shared_ptr<Bee> > bees;

	///	Particles for when PlayerBullets die.
	ParticleGenerator particles;

	///	The biomes in the world.
	Biome biomes[NumBiomes][NumBiomes];

	///	Our sounder.
	shared_ptr<PastarenaSounder> sounder;
};

#endif
