//	PolyphonicSynth.h - Template class to make any monophonic synth polyphonic.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2010  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef POLYPHONICSYNTH_H_
#define POLYPHONICSYNTH_H_

#include "FilePath.h"
#include "UTF8File.h"

#include <string>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <cassert>
#include <iostream>

///	Template class to make any monophonic synth polyphonic.
template<class T>
class PolyphonicSynth
{
  public:
	///	Constructor.
	PolyphonicSynth() {};
	///	Destructor.
	~PolyphonicSynth() {};

	///	Returns the audio for the current block.
	/*!
		\param buffer Stereo (interleaved) buffer to write audio to.
		\param numSamples The number of samples in the current block (i.e.
		buffer will contain 2x numSamples.
	 */
	void getAudio(float *buffer, int numSamples)
	{
		int i;

		for(i=0;i<MaxVoices;++i)
		{
			if(voices[i].getActive())
				voices[i].getAudio(buffer, numSamples);
		}
	};

	///	Triggers a note at the specified frequency.
	void triggerNote(float freq,
					 int sample,
					 float velocity = 1.0f,
					 float pan = 0.5f)
	{
		int i;

		for(i=0;i<MaxVoices;++i)
		{
			if(!voices[i].getActive())
			{
				voices[i].triggerNote(freq, sample, velocity, pan);
				break;
			}
		}
		assert(i<MaxVoices);
	};
	///	Sets the named parameter.
	/*!
		\param name The name of the parameter to change.
		\param value The value to change the parameter to (0->1).
	 */
	void setParameter(const std::string& name, float value, bool direct = false)
	{
		int i;

		for(i=0;i<MaxVoices;++i)
			voices[i].setParameter(name, value, direct);
	};

	///	Saves the synth's parameters to a .inl file.
	void saveParameters(const std::string& name)
	{
		int i;
		std::string tempPath;
		std::wstring tempName;
		std::wstringstream tempstr;
		const int numParams = voices[0].getNumParameters();

		UTF8File::charToWstring(name, tempName);
		tempstr << tempName << L".inl";

		UTF8File::wstringToChar(FilePath::getDataFilesDir().getChild(tempstr.str()).getPath().c_str(),
								tempPath);
		std::ofstream outFile(tempPath);

		outFile << "//\tGeneric synth parameters file generated by the PolyphonicSynth class." << std::endl;
		outFile << "//\t----------------------------------------------------------------------------" << std::endl;
		outFile << "//\tThis file is part of 'GameTemplate', template code for a game." << std::endl;
		outFile << "//\tCopyright (C) 2010  Niall Moody" << std::endl;
		outFile << "//\t" << std::endl;
		outFile << "//\tThis program is free software: you can redistribute it and/or modify" << std::endl;
		outFile << "//\tit under the terms of the GNU General Public License as published by" << std::endl;
		outFile << "//\tthe Free Software Foundation, either version 3 of the License, or" << std::endl;
		outFile << "//\t(at your option) any later version." << std::endl;
		outFile << "//\t" << std::endl;
		outFile << "//\tThis program is distributed in the hope that it will be useful," << std::endl;
		outFile << "//\tbut WITHOUT ANY WARRANTY; without even the implied warranty of" << std::endl;
		outFile << "//\tMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" << std::endl;
		outFile << "//\tGNU General Public License for more details." << std::endl;
		outFile << "//\t" << std::endl;
		outFile << "//\tYou should have received a copy of the GNU General Public License" << std::endl;
		outFile << "//\talong with this program.  If not, see <http://www.gnu.org/licenses/>." << std::endl;
		outFile << "//\t----------------------------------------------------------------------------" << std::endl;
		outFile << std::endl;
		outFile << "namespace " << name << std::endl;
		outFile << "{" << std::endl;
		outFile << std::endl;
		outFile << "const int numParameters = " << numParams << ";" << std::endl;
		outFile << std::endl;
		outFile << "const float parameters[] = {" << std::endl;
		for(i=0;i<numParams;++i)
		{
			outFile << std::fixed << std::setprecision(8) << voices[0].getParameter(i) << "f";
			if(i < (numParams-1))
				outFile << "," << std::endl;
			else
				outFile << std::endl;
		}
		outFile << "};" << std::endl;
		outFile << std::endl;
		outFile << "};" << std::endl;
	};
	///	Loads the synth's parameters from a .inl file.
	void loadParameters(const FilePath& path)
	{
		int i;
		float tempf;
		int numParams;
		std::string tempstr2;
		std::stringstream tempstr;
		std::ifstream inFile(path.getPath().c_str());

		//Ignore comments etc.
		for(i=0;i<22;++i)
			std::getline(inFile, tempstr2);

		//Read num parameters.
		std::getline(inFile, tempstr2);
		tempstr << tempstr2.substr(26, tempstr2.length()-27);
		tempstr >> numParams;
		if(numParams != voices[0].getNumParameters())
		{
			std::cout << "PolyphonicSynth: Number of parameters in file does not match synth's number of parameters." << std::endl;
			return;
		}

		//Ignore in-between lines.
		std::getline(inFile, tempstr2);
		std::getline(inFile, tempstr2);

		i = 0;
		std::getline(inFile, tempstr2);
		while(tempstr2[0] != '}')
		{
			tempstr.clear();
			tempstr.str("");
			tempstr << tempstr2.substr(0, tempstr2.length()-2);
			tempstr >> tempf;
			setParameter(voices[0].getParameterName(i), tempf, true);
			++i;
			std::getline(inFile, tempstr2);
		}
	};
	///	Loads the synth's parameters directly from a .inl file.
	void loadParameters(int numParams, const float params[])
	{
		int i;

		if(numParams != voices[0].getNumParameters())
		{
			std::cout << "PolyphonicSynth: Number of parameters in file does not match synth's number of parameters." << std::endl;
			return;
		}

		for(i=0;i<numParams;++i)
			setParameter(voices[0].getParameterName(i), params[i], true);
	};
  private:
	///	The maximum number of voices.
	enum
	{
		MaxVoices = 64
	};

	///	The voices of the synth.
	T voices[MaxVoices];
};

#endif
